@@ -10,7 +10,7 @@ from .utils import DesignRegion
 
 from .basis import BilinearInterpolationBasis
 
-from .optimization_problem import (OptimizationProblem)
+from .optimization_problem import OptimizationProblem
 
 from .filter_source import FilteredSource
 
@@ -18,6 +18,8 @@ from .filters import *
 
 from .connectivity import *
 
+from .unfilter_design import *
+
 try:
     from .wrapper import MeepJaxWrapper
 except ModuleNotFoundError as _:
