@@ -1,4 +1,4 @@
-/* Copyright (C) 2005-2021 Massachusetts Institute of Technology
+/* Copyright (C) 2005-2024 Massachusetts Institute of Technology
 %
 %  This program is free software; you can redistribute it and/or modify
 %  it under the terms of the GNU General Public License as published by
@@ -20,20 +20,64 @@
    h5fields.cpp
 */
 
+#include <algorithm>
 #include <stdio.h>
 #include <string.h>
 #include <math.h>
 
 #include "meep_internals.hpp"
+#include "config.h"
+
+#define UNUSED(x) (void)x // silence compiler warnings
+
+namespace meep {
+namespace {
 
 using namespace std;
 
-typedef complex<double> cdouble;
+#ifdef HAVE_MPI
+constexpr size_t ARRAY_TO_ALL_BUFSIZE = 1 << 16; // Use (64k * 8 bytes) of buffer
+#endif
 
-namespace meep {
+/***************************************************************/
+/* repeatedly call sum_to_all to consolidate all entries of    */
+/* an array on all cores.                                      */
+/* array: in/out ptr to the data                               */
+/* array_size: data size in multiples of sizeof(realnum)       */
+/***************************************************************/
+realnum *array_to_all(realnum *array, size_t array_size) {
+#ifdef HAVE_MPI
+  realnum *buffer = new realnum[ARRAY_TO_ALL_BUFSIZE];
+  ptrdiff_t offset = 0;
+  size_t remaining = array_size;
+  while (remaining != 0) {
+    size_t xfer_size = (remaining > ARRAY_TO_ALL_BUFSIZE ? ARRAY_TO_ALL_BUFSIZE : remaining);
+    sum_to_all(array + offset, buffer, xfer_size);
+    memcpy(array + offset, buffer, xfer_size * sizeof(realnum));
+    remaining -= xfer_size;
+    offset += xfer_size;
+  }
+  delete[] buffer;
+#else
+  UNUSED(array_size);
+#endif
+  return array;
+}
+
+complex<realnum> *array_to_all(complex<realnum> *array, size_t array_size) {
+  return (complex<realnum> *)array_to_all((realnum *)array, 2 * array_size);
+}
+
+} // namespace
 
 /***************************************************************************/
 
+std::complex<double> cdouble(std::complex<float> z) {
+  return std::complex<double>(real(z), imag(z));
+}
+
+std::complex<double> cdouble(std::complex<double> z) { return z; }
+
 typedef struct {
 
   // information related to the volume covered by the
@@ -62,8 +106,8 @@ typedef struct {
 
   // temporary internal storage buffers
   component *cS;
-  cdouble *ph;
-  cdouble *fields;
+  complex<double> *ph;
+  complex<realnum> *fields;
   ptrdiff_t *offsets;
 
   double frequency;
@@ -80,19 +124,18 @@ typedef struct {
   bool empty_dim[5];
 } array_slice_data;
 
-#define UNUSED(x) (void)x // silence compiler warnings
-
 /* passthrough field function equivalent to component_fun in h5fields.cpp */
-static cdouble default_field_func(const cdouble *fields, const vec &loc, void *data_) {
+static complex<double> default_field_func(const complex<realnum> *fields, const vec &loc,
+                                          void *data_) {
   (void)loc;   // unused
   (void)data_; // unused
-  return fields[0];
+  return cdouble(fields[0]);
 }
 
-static double default_field_rfunc(const cdouble *fields, const vec &loc, void *data_) {
+static double default_field_rfunc(const complex<realnum> *fields, const vec &loc, void *data_) {
   (void)loc;   // unused
   (void)data_; // unused
-  return real(fields[0]);
+  return real(cdouble(fields[0]));
 }
 
 /***************************************************************/
@@ -138,7 +181,7 @@ static void get_array_slice_dimensions_chunkloop(fields_chunk *fc, int ichnk, co
 typedef struct {
   component source_component;
   ivec slice_imin, slice_imax;
-  cdouble *slice;
+  complex<realnum> *slice;
 } source_slice_data;
 
 bool in_range(int imin, int i, int imax) { return (imin <= i && i <= imax); }
@@ -176,18 +219,18 @@ static void get_source_slice_chunkloop(fields_chunk *fc, int ichnk, component cg
   if (has_direction(dim, Y)) NY = ((slice_imax - slice_imin).in_direction(Y) / 2) + 1;
 
   for (int ft = 0; ft < NUM_FIELD_TYPES; ft++)
-    for (src_vol *s = fc->sources[ft]; s; s = s->next) {
+    for (const src_vol &s : fc->get_sources(static_cast<field_type>(ft))) {
       component cS = S.transform(data->source_component, -sn);
-      if (s->c != cS) continue;
+      if (s.c != cS) continue;
 
       // loop over point sources in this src_vol. for each point source,
       // the src_vol stores the amplitude and the global index of the
       // symmetry-parent grid point, from which we need to compute the
       // local index of the symmetry-child grid point within this
       // slice (that is, if it even lies within the slice)
-      for (size_t npt = 0; npt < s->npts; npt++) {
-        cdouble amp = s->A[npt];
-        ptrdiff_t chunk_index = s->index[npt];
+      for (size_t npt = 0; npt < s.num_points(); npt++) {
+        const complex<double> &amp = s.amplitude_at(npt);
+        ptrdiff_t chunk_index = s.index_at(npt);
         ivec iloc_parent = fc->gv.iloc(Dielectric, chunk_index);
         ivec iloc_child = S.transform(iloc_parent, sn) + shift;
         if (!in_subgrid(slice_imin, iloc_child, slice_imax)) continue; // source point outside slice
@@ -238,13 +281,11 @@ static void get_array_slice_chunkloop(fields_chunk *fc, int ichnk, component cgr
   LOOP_OVER_DIRECTIONS(permute.dim, d) { permute.set_direction(d, abs(permute.in_direction(d))); }
 
   // compute the size of the chunk to output, and its strides etc.
-  size_t slice_size = 1;
   for (int i = 0; i < data->rank; ++i) {
     direction d = data->ds[i];
     int isd = isS.in_direction(d), ied = ieS.in_direction(d);
-    start[i] = (min(isd, ied) - data->min_corner.in_direction(d)) / 2;
+    start[i] = (std::min(isd, ied) - data->min_corner.in_direction(d)) / 2;
     count[i] = abs(ied - isd) / 2 + 1;
-    slice_size *= count[i];
     if (ied < isd) offset[permute.in_direction(d)] = count[i] - 1;
   }
 
@@ -274,18 +315,19 @@ static void get_array_slice_chunkloop(fields_chunk *fc, int ichnk, component cgr
   // Otherwise proceed to compute the function of field components to be   //
   // tabulated on the slice, exactly as in fields::integrate.              //
   //-----------------------------------------------------------------------//
-  double *slice = 0;
-  cdouble *zslice = 0;
+  realnum *slice = 0;
+  complex<realnum> *zslice = 0;
   bool complex_data = (data->rfun == 0);
   if (complex_data)
-    zslice = (cdouble *)data->vslice;
+    zslice = (complex<realnum> *)data->vslice;
   else
-    slice = (double *)data->vslice;
+    slice = (realnum *)data->vslice;
 
   ptrdiff_t *off = data->offsets;
   component *cS = data->cS;
   double frequency = data->frequency;
-  complex<double> *fields = data->fields, *ph = data->ph;
+  complex<realnum> *fields = data->fields;
+  complex<double> *ph = data->ph;
   const component *iecs = data->inveps_cs;
   const direction *ieds = data->inveps_ds;
   ptrdiff_t ieos[6];
@@ -390,30 +432,6 @@ static void get_array_slice_chunkloop(fields_chunk *fc, int ichnk, component cgr
   } // LOOP_OVER_IVECS
 }
 
-/***************************************************************/
-/* repeatedly call sum_to_all to consolidate all entries of    */
-/* an array on all cores.                                      */
-/***************************************************************/
-#define BUFSIZE 1 << 16 // use 64k buffer
-double *array_to_all(double *array, size_t array_size) {
-  double *buffer = new double[BUFSIZE];
-  ptrdiff_t offset = 0;
-  size_t remaining = array_size;
-  while (remaining != 0) {
-    size_t xfer_size = (remaining > BUFSIZE ? BUFSIZE : remaining);
-    sum_to_all(array + offset, buffer, xfer_size);
-    memcpy(array + offset, buffer, xfer_size * sizeof(double));
-    remaining -= xfer_size;
-    offset += xfer_size;
-  }
-  delete[] buffer;
-  return array;
-}
-
-cdouble *array_to_all(cdouble *array, size_t array_size) {
-  return (cdouble *)array_to_all((double *)array, 2 * array_size);
-}
-
 /***************************************************************/
 /* given a volume, fill in the dims[] and dirs[] arrays        */
 /* describing the array slice needed to store field data for   */
@@ -455,7 +473,7 @@ int fields::get_array_slice_dimensions(const volume &where, size_t dims[3], dire
   loop_in_chunks(get_array_slice_dimensions_chunkloop, (void *)data, where, cgrid, use_symmetry,
                  snap_empty_dimensions);
 
-  am_now_working_on(MpiTime);
+  am_now_working_on(MpiAllTime);
   data->min_corner = -max_to_all(-data->min_corner); // i.e., min_to_all
   data->max_corner = max_to_all(data->max_corner);
   if (min_max_loc) LOOP_OVER_DIRECTIONS(gv.dim, d) {
@@ -470,7 +488,7 @@ int fields::get_array_slice_dimensions(const volume &where, size_t dims[3], dire
   int rank = 0;
   size_t slice_size = 1;
   LOOP_OVER_DIRECTIONS(gv.dim, d) {
-    if (rank >= 3) abort("too many dimensions in array_slice");
+    if (rank >= 3) meep::abort("too many dimensions in array_slice");
     size_t n = (data->max_corner.in_direction(d) - data->min_corner.in_direction(d)) / 2 + 1;
     if (where.in_direction(d) == 0.0 && collapse_empty_dimensions) n = 1;
     if (n > 1) {
@@ -502,55 +520,68 @@ bool increment(size_t *n, size_t *nMax, int rank) {
   return true;
 }
 
+// get the dimensions of reduced arrays (i.e., arrays collapsed along empty dimensions of original
+// volume)
+void reduce_array_dimensions(volume where, int rank, size_t dims[3], direction dirs[3],
+                             size_t stride[3], int &reduced_rank, size_t reduced_dims[3],
+                             direction reduced_dirs[3], size_t reduced_stride[3]) {
+
+  reduced_rank = 0;
+  reduced_dims[0] = reduced_dims[1] = reduced_dims[2] = stride[0] = stride[1] = stride[2] =
+      reduced_stride[0] = reduced_stride[1] = reduced_stride[2] = 1;
+  for (int r = 0; r < rank; r++) {
+    if (where.in_direction(dirs[r]) == 0.0)
+      reduced_stride[r] = 0; // degenerate dimension, to be collapsed
+    else {
+      reduced_dirs[reduced_rank] = dirs[r];
+      reduced_dims[reduced_rank++] =
+          dims[r]; // reduced_dims is the size of the array after collapsing
+    }
+  }
+  /*--------------------------------------------------------------*/
+  /*- set up strides into full and reduced arrays                 */
+  /*--------------------------------------------------------------*/
+  if (rank == 2)
+    stride[0] = dims[1]; // rstride is already all set in this case
+  else if (rank == 3) {
+    stride[0] = dims[1] * dims[2];
+    stride[1] = dims[2];
+    if (reduced_rank == 2) reduced_stride[reduced_stride[0] != 0 ? 0 : 1] = reduced_dims[1];
+  }
+}
+
 // data_size = 1,2 for real,complex-valued array
-double *collapse_array(double *array, int *rank, size_t dims[3], direction dirs[3], volume where,
-                       int data_size = 1) {
+realnum *collapse_array(realnum *array, int *rank, size_t dims[3], direction dirs[3], volume where,
+                        int data_size = 1) {
 
   /*--------------------------------------------------------------*/
   /*- detect empty dimensions and compute rank and strides for    */
   /*- collapsed array                                             */
   /*--------------------------------------------------------------*/
   int full_rank = *rank;
-  if (full_rank == 0) return array;
-
-  int reduced_rank = 0;
-  size_t reduced_dims[3], reduced_stride[3] = {1, 1, 1};
+  size_t reduced_dims[3], reduced_stride[3], stride[3]; // the latter for non-reduced array strides
   direction reduced_dirs[3];
-  for (int r = 0; r < full_rank; r++) {
-    if (where.in_direction(dirs[r]) == 0.0)
-      reduced_stride[r] = 0; // degenerate dimension, to be collapsed
-    else {
-      reduced_dirs[reduced_rank] = dirs[r];
-      reduced_dims[reduced_rank++] = dims[r];
-    }
-  }
-  if (reduced_rank==0) {
+  size_t reduced_grid_size;
+  int reduced_rank;
+  reduce_array_dimensions(where, full_rank, dims, dirs, stride, reduced_rank, reduced_dims,
+                          reduced_dirs, reduced_stride);
+
+  if (full_rank == 0) return array;
+  if (reduced_rank == 0) {
     *rank = 0;
     return array; // return array as is for singleton use case
   }
   if (reduced_rank == full_rank) return array; // nothing to collapse
 
-  /*--------------------------------------------------------------*/
-  /*- set up strides into full and reduced arrays                 */
-  /*--------------------------------------------------------------*/
-  size_t stride[3] = {1, 1, 1}; // non-reduced array strides
-  if (full_rank == 2)
-    stride[0] = dims[1]; // rstride is already all set in this case
-  else if (full_rank == 3) {
-    stride[0] = dims[1] * dims[2];
-    stride[1] = dims[2];
-    if (reduced_rank == 2)
-      reduced_stride[reduced_stride[0] != 0 ? 0 : 1] = reduced_dims[1];
-  }
-
   /*--------------------------------------------------------------*/
   /*- allocate reduced array and compress full array into it     -*/
   /*--------------------------------------------------------------*/
-  size_t reduced_grid_size = reduced_dims[0] * (reduced_rank == 2 ? reduced_dims[1] : 1);
+  reduced_grid_size = reduced_dims[0] * reduced_dims[1] * reduced_dims[2];
   size_t reduced_array_size = data_size * reduced_grid_size;
-  double *reduced_array = new double[reduced_array_size];
-  if (!reduced_array) abort("%s:%i: out of memory (%zu)", __FILE__, __LINE__, reduced_array_size);
-  memset(reduced_array, 0, reduced_array_size * sizeof(double));
+  realnum *reduced_array = new realnum[reduced_array_size];
+  if (!reduced_array)
+    meep::abort("%s:%i: out of memory (%zu)", __FILE__, __LINE__, reduced_array_size);
+  memset(reduced_array, 0, reduced_array_size * sizeof(realnum));
 
   size_t n[3] = {0, 0, 0};
   do {
@@ -569,9 +600,9 @@ double *collapse_array(double *array, int *rank, size_t dims[3], direction dirs[
   return reduced_array;
 }
 
-cdouble *collapse_array(cdouble *array, int *rank, size_t dims[3], direction dirs[3],
-                        volume where) {
-  return (cdouble *)collapse_array((double *)array, rank, dims, dirs, where, 2);
+complex<realnum> *collapse_array(complex<realnum> *array, int *rank, size_t dims[3],
+                                 direction dirs[3], volume where) {
+  return (complex<realnum> *)collapse_array((realnum *)array, rank, dims, dirs, where, 2);
 }
 
 /**********************************************************************/
@@ -595,7 +626,8 @@ void *fields::do_get_array_slice(const volume &where, std::vector<component> com
   int elem_size = complex_data ? 2 : 1;
   void *vslice_uncollapsed;
 
-  vslice_uncollapsed = memset(new double[slice_size * elem_size], 0, slice_size * elem_size * sizeof(double));
+  vslice_uncollapsed =
+      memset(new realnum[slice_size * elem_size], 0, slice_size * elem_size * sizeof(realnum));
 
   data.vslice = vslice_uncollapsed;
   data.snap = snap;
@@ -606,8 +638,8 @@ void *fields::do_get_array_slice(const volume &where, std::vector<component> com
   data.frequency = frequency;
   int num_components = components.size();
   data.cS = new component[num_components];
-  data.ph = new cdouble[num_components];
-  data.fields = new cdouble[num_components];
+  data.ph = new complex<double>[num_components];
+  data.fields = new complex<realnum>[num_components];
   data.offsets = new ptrdiff_t[2 * num_components];
   memset(data.offsets, 0, 2 * num_components * sizeof(ptrdiff_t));
   data.empty_dim[0] = data.empty_dim[1] = data.empty_dim[2] = data.empty_dim[3] =
@@ -623,7 +655,7 @@ void *fields::do_get_array_slice(const volume &where, std::vector<component> com
       break;
     }
   if (needs_dielectric) FOR_ELECTRIC_COMPONENTS(c) if (gv.has_field(c)) {
-      if (data.ninveps == 3) abort("more than 3 field components??");
+      if (data.ninveps == 3) meep::abort("more than 3 field components??");
       data.inveps_cs[data.ninveps] = c;
       data.inveps_ds[data.ninveps] = component_direction(c);
       ++data.ninveps;
@@ -638,7 +670,7 @@ void *fields::do_get_array_slice(const volume &where, std::vector<component> com
       break;
     }
   if (needs_permeability) FOR_MAGNETIC_COMPONENTS(c) if (gv.has_field(c)) {
-      if (data.ninvmu == 3) abort("more than 3 field components??");
+      if (data.ninvmu == 3) meep::abort("more than 3 field components??");
       data.invmu_cs[data.ninvmu] = c;
       data.invmu_ds[data.ninvmu] = component_direction(c);
       ++data.ninvmu;
@@ -647,19 +679,20 @@ void *fields::do_get_array_slice(const volume &where, std::vector<component> com
   loop_in_chunks(get_array_slice_chunkloop, (void *)&data, where, Centered, true, snap);
 
   if (!snap) {
-    double *slice = collapse_array((double *)vslice_uncollapsed, &rank, dims, dirs, where, elem_size);
+    realnum *slice =
+        collapse_array((realnum *)vslice_uncollapsed, &rank, dims, dirs, where, elem_size);
     rank = get_array_slice_dimensions(where, dims, dirs, true, false, 0, &data);
     slice_size = data.slice_size;
-    vslice_uncollapsed = (double*) slice;
+    vslice_uncollapsed = (realnum *)slice;
   }
   if (vslice) {
-    memcpy(vslice, vslice_uncollapsed, slice_size * elem_size * sizeof(double));
-    delete[] (double*) vslice_uncollapsed;
+    memcpy(vslice, vslice_uncollapsed, slice_size * elem_size * sizeof(realnum));
+    delete[] (realnum *)vslice_uncollapsed;
   }
   else
     vslice = vslice_uncollapsed;
 
-  array_to_all((double *)vslice, elem_size * slice_size);
+  array_to_all((realnum *)vslice, elem_size * slice_size);
 
   delete[] data.offsets;
   delete[] data.fields;
@@ -673,48 +706,51 @@ void *fields::do_get_array_slice(const volume &where, std::vector<component> com
 /***************************************************************/
 /* entry points to get_array_slice                             */
 /***************************************************************/
-double *fields::get_array_slice(const volume &where, std::vector<component> components,
-                                field_rfunction rfun, void *fun_data, double *slice,
-                                double frequency, bool snap) {
-  return (double *)do_get_array_slice(where, components, 0, rfun, fun_data, (void *)slice,
-                                      frequency, snap);
-}
-
-cdouble *fields::get_complex_array_slice(const volume &where, std::vector<component> components,
-                                         field_function fun, void *fun_data, cdouble *slice,
-                                         double frequency, bool snap) {
-  return (cdouble *)do_get_array_slice(where, components, fun, 0, fun_data, (void *)slice,
+realnum *fields::get_array_slice(const volume &where, std::vector<component> components,
+                                 field_rfunction rfun, void *fun_data, realnum *slice,
+                                 double frequency, bool snap) {
+  return (realnum *)do_get_array_slice(where, components, 0, rfun, fun_data, (void *)slice,
                                        frequency, snap);
 }
 
-double *fields::get_array_slice(const volume &where, component c, double *slice, double frequency,
-                                bool snap) {
-  std::vector<component> components(1);
-  components[0] = c;
-  return (double *)do_get_array_slice(where, components, 0, default_field_rfunc, 0, (void *)slice,
-                                      frequency, snap);
+complex<realnum> *fields::get_complex_array_slice(const volume &where,
+                                                  std::vector<component> components,
+                                                  field_function fun, void *fun_data,
+                                                  complex<realnum> *slice, double frequency,
+                                                  bool snap) {
+  return (complex<realnum> *)do_get_array_slice(where, components, fun, 0, fun_data, (void *)slice,
+                                                frequency, snap);
 }
 
-double *fields::get_array_slice(const volume &where, derived_component c, double *slice,
-                                double frequency, bool snap) {
+realnum *fields::get_array_slice(const volume &where, component c, realnum *slice, double frequency,
+                                 bool snap) {
+  std::vector<component> components(1);
+  components[0] = c;
+  return (realnum *)do_get_array_slice(where, components, 0, default_field_rfunc, 0, (void *)slice,
+                                       frequency, snap);
+}
+
+realnum *fields::get_array_slice(const volume &where, derived_component c, realnum *slice,
+                                 double frequency, bool snap) {
   int nfields;
   component carray[12];
   field_rfunction rfun = derived_component_func(c, gv, nfields, carray);
   std::vector<component> cs(carray, carray + nfields);
-  return (double *)do_get_array_slice(where, cs, 0, rfun, &nfields, (void *)slice,
-                                      frequency, snap);
+  return (realnum *)do_get_array_slice(where, cs, 0, rfun, &nfields, (void *)slice, frequency,
+                                       snap);
 }
 
-cdouble *fields::get_complex_array_slice(const volume &where, component c, cdouble *slice,
-                                         double frequency, bool snap) {
+complex<realnum> *fields::get_complex_array_slice(const volume &where, component c,
+                                                  complex<realnum> *slice, double frequency,
+                                                  bool snap) {
   std::vector<component> components(1);
   components[0] = c;
-  return (cdouble *)do_get_array_slice(where, components, default_field_func, 0, 0, (void *)slice,
-                                       frequency, snap);
+  return (complex<realnum> *)do_get_array_slice(where, components, default_field_func, 0, 0,
+                                                (void *)slice, frequency, snap);
 }
 
-cdouble *fields::get_source_slice(const volume &where, component source_slice_component,
-                                  cdouble *slice) {
+complex<realnum> *fields::get_source_slice(const volume &where, component source_slice_component,
+                                           complex<realnum> *slice) {
   size_t dims[3];
   direction dirs[3];
   vec min_max_loc[2];
@@ -725,18 +761,18 @@ cdouble *fields::get_source_slice(const volume &where, component source_slice_co
   data.source_component = source_slice_component;
   data.slice_imin = gv.round_vec(min_max_loc[0]);
   data.slice_imax = gv.round_vec(min_max_loc[1]);
-  data.slice = new cdouble[slice_size];
-  if (!data.slice) abort("%s:%i: out of memory (%zu)", __FILE__, __LINE__, slice_size);
+  data.slice = new complex<realnum>[slice_size];
+  if (!data.slice) meep::abort("%s:%i: out of memory (%zu)", __FILE__, __LINE__, slice_size);
 
   loop_in_chunks(get_source_slice_chunkloop, (void *)&data, where, Centered, true, false);
 
-  cdouble *slice_collapsed = collapse_array(data.slice, &rank, dims, dirs, where);
+  complex<realnum> *slice_collapsed = collapse_array(data.slice, &rank, dims, dirs, where);
   rank = get_array_slice_dimensions(where, dims, dirs, true, false);
   slice_size = dims[0] * (rank >= 2 ? dims[1] : 1) * (rank == 3 ? dims[2] : 1);
 
   if (slice) {
-    memcpy(slice, slice_collapsed, 2 * slice_size * sizeof(double));
-    delete[] (cdouble*) slice_collapsed;
+    memcpy(slice, slice_collapsed, 2 * slice_size * sizeof(realnum));
+    delete[] (complex<realnum> *)slice_collapsed;
   }
   else
     slice = slice_collapsed;
@@ -750,26 +786,22 @@ cdouble *fields::get_source_slice(const volume &where, component source_slice_co
 /***************************************************************/
 /***************************************************************/
 std::vector<double> fields::get_array_metadata(const volume &where) {
+  if (where.dim == Dcyl)
+    meep::abort("get_array_metadata does not support cylindrical coordinates.");
 
   /* get extremal corners of subgrid and array of weights, collapsed if necessary */
   size_t dims[3];
   direction dirs[3];
   vec min_max_loc[2]; // extremal points in subgrid
-  int rank = get_array_slice_dimensions(where, dims, dirs, true, false, min_max_loc);
+  get_array_slice_dimensions(where, dims, dirs, true, false, min_max_loc);
 
-  int full_rank = rank;
-  direction full_dirs[3];
-  for (int fr = 0; fr < rank; fr++)
-    full_dirs[fr] = dirs[fr];
-
-  double *weights = get_array_slice(where, NO_COMPONENT);
+  realnum *weights = get_array_slice(where, NO_COMPONENT);
 
   /* get length and endpoints of x,y,z tics arrays */
   size_t nxyz[3] = {1, 1, 1};
   double xyzmin[3] = {0.0, 0.0, 0.0}, xyzmax[3] = {0.0, 0.0, 0.0};
-  for (int fr = 0, rr = 0; fr < full_rank; fr++) {
-    direction d = full_dirs[fr];
-    int nd = d - X;
+  for (int nd = 0, rr = 0; nd < 3; ++nd) {
+    direction d = direction(nd);
     if (where.in_direction(d) == 0.0) {
       xyzmin[nd] = xyzmax[nd] = where.in_direction_min(d);
       nxyz[nd] = 1;
