@@ -1,4 +1,4 @@
-/* Copyright (C) 2005-2019 Massachusetts Institute of Technology
+/* Copyright (C) 2005-2024 Massachusetts Institute of Technology
 %
 %  This program is free software; you can redistribute it and/or modify
 %  it under the terms of the GNU General Public License as published by
@@ -36,7 +36,7 @@ typedef struct {
   h5file *file;
   ivec min_corner, max_corner;
   int num_chunks;
-  realnum *buf;
+  double *buf;
   size_t bufsz;
   int rank;
   direction ds[3];
@@ -48,9 +48,9 @@ typedef struct {
   const component *components;
   component *cS;
   complex<double> *ph;
-  complex<double> *fields;
+  complex<realnum> *fields;
   ptrdiff_t *offsets;
-  double omega;
+  double frequency;
   int ninveps;
   component inveps_cs[3];
   direction inveps_ds[3];
@@ -86,7 +86,7 @@ static void h5_findsize_chunkloop(fields_chunk *fc, int ichnk, component cgrid,
   LOOP_OVER_DIRECTIONS(fc->gv.dim, d) {
     bufsz *= (ie.in_direction(d) - is.in_direction(d)) / 2 + 1;
   }
-  data->bufsz = max(data->bufsz, bufsz);
+  data->bufsz = std::max(data->bufsz, bufsz);
 }
 
 static void h5_output_chunkloop(fields_chunk *fc, int ichnk, component cgrid, ivec is, ivec ie,
@@ -124,7 +124,7 @@ static void h5_output_chunkloop(fields_chunk *fc, int ichnk, component cgrid, iv
   for (int i = 0; i < data->rank; ++i) {
     direction d = data->ds[i];
     int isd = isS.in_direction(d), ied = ieS.in_direction(d);
-    start[i] = (min(isd, ied) - data->min_corner.in_direction(d)) / 2;
+    start[i] = (std::min(isd, ied) - data->min_corner.in_direction(d)) / 2;
     count[i] = abs(ied - isd) / 2 + 1;
     if (ied < isd) offset[permute.in_direction(d)] = count[i] - 1;
   }
@@ -143,8 +143,9 @@ static void h5_output_chunkloop(fields_chunk *fc, int ichnk, component cgrid, iv
 
   ptrdiff_t *off = data->offsets;
   component *cS = data->cS;
-  complex<double> *fields = data->fields, *ph = data->ph;
-  double omega = data->omega;
+  complex<realnum> *fields = data->fields;
+  complex<double> *ph = data->ph;
+  double frequency = data->frequency;
   const component *iecs = data->inveps_cs;
   const direction *ieds = data->inveps_ds;
   ptrdiff_t ieos[6];
@@ -173,28 +174,28 @@ static void h5_output_chunkloop(fields_chunk *fc, int ichnk, component cgrid, iv
 
     for (int i = 0; i < data->num_fields; ++i) {
       if (cS[i] == Dielectric) {
-        double tr = 0.0;
+        complex<double> tr(0.0, 0.0);
         for (int k = 0; k < data->ninveps; ++k) {
-          tr += (fc->s->get_chi1inv_at_pt(iecs[k], ieds[k], idx, omega) +
-                 fc->s->get_chi1inv_at_pt(iecs[k], ieds[k], idx + ieos[2 * k], omega) +
-                 fc->s->get_chi1inv_at_pt(iecs[k], ieds[k], idx + ieos[1 + 2 * k], omega) +
+          tr += (fc->s->get_chi1inv_at_pt(iecs[k], ieds[k], idx, frequency) +
+                 fc->s->get_chi1inv_at_pt(iecs[k], ieds[k], idx + ieos[2 * k], frequency) +
+                 fc->s->get_chi1inv_at_pt(iecs[k], ieds[k], idx + ieos[1 + 2 * k], frequency) +
                  fc->s->get_chi1inv_at_pt(iecs[k], ieds[k], idx + ieos[2 * k] + ieos[1 + 2 * k],
-                                          omega));
+                                          frequency));
           if (tr == 0.0) tr += 4.0; // default inveps == 1
         }
-        fields[i] = (4 * data->ninveps) / tr;
+        fields[i] = (4.0 * data->ninveps) / tr;
       }
       else if (cS[i] == Permeability) {
-        double tr = 0.0;
+        complex<double> tr(0.0, 0.0);
         for (int k = 0; k < data->ninvmu; ++k) {
-          tr += (fc->s->get_chi1inv_at_pt(imcs[k], imds[k], idx, omega) +
-                 fc->s->get_chi1inv_at_pt(imcs[k], imds[k], idx + imos[2 * k], omega) +
-                 fc->s->get_chi1inv_at_pt(imcs[k], imds[k], idx + imos[1 + 2 * k], omega) +
+          tr += (fc->s->get_chi1inv_at_pt(imcs[k], imds[k], idx, frequency) +
+                 fc->s->get_chi1inv_at_pt(imcs[k], imds[k], idx + imos[2 * k], frequency) +
+                 fc->s->get_chi1inv_at_pt(imcs[k], imds[k], idx + imos[1 + 2 * k], frequency) +
                  fc->s->get_chi1inv_at_pt(imcs[k], imds[k], idx + imos[2 * k] + imos[1 + 2 * k],
-                                          omega));
+                                          frequency));
           if (tr == 0.0) tr += 4.0; // default invmu == 1
         }
-        fields[i] = (4 * data->ninvmu) / tr;
+        fields[i] = (4.0 * data->ninvmu) / tr;
       }
       else {
         double f[2];
@@ -224,7 +225,7 @@ static void h5_output_chunkloop(fields_chunk *fc, int ichnk, component cgrid, iv
 void fields::output_hdf5(h5file *file, const char *dataname, int num_fields,
                          const component *components, field_function fun, void *fun_data_, int reim,
                          const volume &where, bool append_data, bool single_precision,
-                         double omega) {
+                         double frequency) {
   am_now_working_on(FieldOutput);
   h5_output_data data;
 
@@ -238,15 +239,17 @@ void fields::output_hdf5(h5file *file, const char *dataname, int num_fields,
   loop_in_chunks(h5_findsize_chunkloop, (void *)&data, where, Centered, true, true);
 
   file->prevent_deadlock(); // can't hold a lock since *_to_all is collective
+  am_now_working_on(MpiAllTime);
   data.max_corner = max_to_all(data.max_corner);
   data.min_corner = -max_to_all(-data.min_corner); // i.e., min_to_all
   data.num_chunks = sum_to_all(data.num_chunks);
+  finished_working();
   if (data.num_chunks == 0 || !(data.min_corner <= data.max_corner)) return; // no data to write;
 
   int rank = 0;
   size_t dims[3];
   LOOP_OVER_DIRECTIONS(gv.dim, d) {
-    if (rank >= 3) abort("too many dimensions in output_hdf5");
+    if (rank >= 3) meep::abort("too many dimensions in output_hdf5");
     size_t n =
         std::max(0, (data.max_corner.in_direction(d) - data.min_corner.in_direction(d)) / 2 + 1);
 
@@ -259,18 +262,18 @@ void fields::output_hdf5(h5file *file, const char *dataname, int num_fields,
 
   file->create_or_extend_data(dataname, rank, dims, append_data, single_precision);
 
-  data.buf = new realnum[data.bufsz];
+  data.buf = new double[data.bufsz];
 
   data.num_fields = num_fields;
   data.components = components;
   data.cS = new component[num_fields];
   data.ph = new complex<double>[num_fields];
-  data.fields = new complex<double>[num_fields];
+  data.fields = new complex<realnum>[num_fields];
   data.fun = fun;
   data.fun_data_ = fun_data_;
 
   /* compute inverse-epsilon directions for computing Dielectric fields */
-  data.omega = omega;
+  data.frequency = frequency;
   data.ninveps = 0;
   bool needs_dielectric = false;
   for (int i = 0; i < num_fields; ++i)
@@ -279,7 +282,7 @@ void fields::output_hdf5(h5file *file, const char *dataname, int num_fields,
       break;
     }
   if (needs_dielectric) FOR_ELECTRIC_COMPONENTS(c) if (gv.has_field(c)) {
-      if (data.ninveps == 3) abort("more than 3 field components??");
+      if (data.ninveps == 3) meep::abort("more than 3 field components??");
       data.inveps_cs[data.ninveps] = c;
       data.inveps_ds[data.ninveps] = component_direction(c);
       ++data.ninveps;
@@ -294,7 +297,7 @@ void fields::output_hdf5(h5file *file, const char *dataname, int num_fields,
       break;
     }
   if (needs_permeability) FOR_MAGNETIC_COMPONENTS(c) if (gv.has_field(c)) {
-      if (data.ninvmu == 3) abort("more than 3 field components??");
+      if (data.ninvmu == 3) meep::abort("more than 3 field components??");
       data.invmu_cs[data.ninvmu] = c;
       data.invmu_ds[data.ninvmu] = component_direction(c);
       ++data.ninvmu;
@@ -320,23 +323,23 @@ void fields::output_hdf5(h5file *file, const char *dataname, int num_fields,
 void fields::output_hdf5(const char *dataname, int num_fields, const component *components,
                          field_function fun, void *fun_data_, const volume &where, h5file *file,
                          bool append_data, bool single_precision, const char *prefix,
-                         bool real_part_only, double omega) {
+                         bool real_part_only, double frequency) {
   bool delete_file;
   if ((delete_file = !file)) file = open_h5file(dataname, h5file::WRITE, prefix, true);
 
   if (real_part_only) {
     output_hdf5(file, dataname, num_fields, components, fun, fun_data_, 0, where, append_data,
-                single_precision, omega);
+                single_precision, frequency);
   }
   else {
     int len = strlen(dataname) + 5;
     char *dataname2 = new char[len];
     snprintf(dataname2, len, "%s%s", dataname, ".r");
     output_hdf5(file, dataname2, num_fields, components, fun, fun_data_, 0, where, append_data,
-                single_precision, omega);
+                single_precision, frequency);
     snprintf(dataname2, len, "%s%s", dataname, ".i");
     output_hdf5(file, dataname2, num_fields, components, fun, fun_data_, 1, where, append_data,
-                single_precision, omega);
+                single_precision, frequency);
     delete[] dataname2;
   }
   if (delete_file) delete file;
@@ -349,7 +352,7 @@ typedef struct {
   void *fun_data_;
 } rintegrand_data;
 
-static complex<double> rintegrand_fun(const complex<double> *fields, const vec &loc, void *data_) {
+static complex<double> rintegrand_fun(const complex<realnum> *fields, const vec &loc, void *data_) {
   rintegrand_data *data = (rintegrand_data *)data_;
   return data->fun(fields, loc, data->fun_data_);
 }
@@ -357,7 +360,7 @@ static complex<double> rintegrand_fun(const complex<double> *fields, const vec &
 void fields::output_hdf5(const char *dataname, int num_fields, const component *components,
                          field_rfunction fun, void *fun_data_, const volume &where, h5file *file,
                          bool append_data, bool single_precision, const char *prefix,
-                         double omega) {
+                         double frequency) {
   bool delete_file;
   if ((delete_file = !file)) file = open_h5file(dataname, h5file::WRITE, prefix, true);
 
@@ -365,41 +368,42 @@ void fields::output_hdf5(const char *dataname, int num_fields, const component *
   data.fun = fun;
   data.fun_data_ = fun_data_;
   output_hdf5(file, dataname, num_fields, components, rintegrand_fun, (void *)&data, 0, where,
-              append_data, single_precision, omega);
+              append_data, single_precision, frequency);
 
   if (delete_file) delete file;
 }
 
 /***************************************************************************/
 
-static complex<double> component_fun(const complex<double> *fields, const vec &loc, void *data_) {
+static complex<double> component_fun(const complex<realnum> *fields, const vec &loc, void *data_) {
   (void)loc;   // unused
   (void)data_; // unused
-  return fields[0];
+  return cdouble(fields[0]);
 }
 
 void fields::output_hdf5(component c, const volume &where, h5file *file, bool append_data,
-                         bool single_precision, const char *prefix, double omega) {
+                         bool single_precision, const char *prefix, double frequency) {
   if (is_derived(int(c))) {
-    output_hdf5(derived_component(c), where, file, append_data, single_precision, prefix, omega);
+    output_hdf5(derived_component(c), where, file, append_data, single_precision, prefix,
+                frequency);
     return;
   }
 
   if (coordinate_mismatch(gv.dim, c)) return;
 
   char dataname[256];
-  bool has_imag = !is_real && c != Dielectric && c != Permeability;
+  bool has_imag = frequency != 0 || (!is_real && c != Dielectric && c != Permeability);
 
   bool delete_file;
   if ((delete_file = !file)) file = open_h5file(component_name(c), h5file::WRITE, prefix, true);
 
   snprintf(dataname, 256, "%s%s", component_name(c), has_imag ? ".r" : "");
   output_hdf5(file, dataname, 1, &c, component_fun, 0, 0, where, append_data, single_precision,
-              omega);
+              frequency);
   if (has_imag) {
     snprintf(dataname, 256, "%s.i", component_name(c));
     output_hdf5(file, dataname, 1, &c, component_fun, 0, 1, where, append_data, single_precision,
-                omega);
+                frequency);
   }
 
   if (delete_file) delete file;
@@ -408,9 +412,9 @@ void fields::output_hdf5(component c, const volume &where, h5file *file, bool ap
 /***************************************************************************/
 
 void fields::output_hdf5(derived_component c, const volume &where, h5file *file, bool append_data,
-                         bool single_precision, const char *prefix, double omega) {
+                         bool single_precision, const char *prefix, double frequency) {
   if (!is_derived(int(c))) {
-    output_hdf5(component(c), where, file, append_data, single_precision, prefix, omega);
+    output_hdf5(component(c), where, file, append_data, single_precision, prefix, frequency);
     return;
   }
 
@@ -421,7 +425,7 @@ void fields::output_hdf5(derived_component c, const volume &where, h5file *file,
   field_rfunction fun = derived_component_func(c, gv, nfields, cs);
 
   output_hdf5(component_name(c), nfields, cs, fun, &nfields, where, file, append_data,
-              single_precision, prefix, omega);
+              single_precision, prefix, frequency);
 }
 
 /***************************************************************************/
@@ -451,8 +455,9 @@ const char *fields::h5file_name(const char *name, const char *prefix, bool times
 h5file *fields::open_h5file(const char *name, h5file::access_mode mode, const char *prefix,
                             bool timestamp) {
   const char *filename = h5file_name(name, prefix, timestamp);
-  if (verbosity > 0 && mode == h5file::WRITE) master_printf("creating output file \"%s\"...\n", filename);
+  if (verbosity > 0 && mode == h5file::WRITE)
+    master_printf("creating output file \"%s\"...\n", filename);
   return new h5file(filename, mode, true);
 }
 
-} // namespace meep
\ No newline at end of file
+} // namespace meep
