@@ -1,27 +1,26 @@
+/* Copyright (C) 2005-2024 Massachusetts Institute of Technology
+%
+%  This program is free software; you can redistribute it and/or modify
+%  it under the terms of the GNU General Public License as published by
+%  the Free Software Foundation; either version 2, or (at your option)
+%  any later version.
+%
+%  This program is distributed in the hope that it will be useful,
+%  but WITHOUT ANY WARRANTY; without even the implied warranty of
+%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+%  GNU General Public License for more details.
+%
+%  You should have received a copy of the GNU General Public License
+%  along with this program; if not, write to the Free Software Foundation,
+%  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+*/
+
 #include "meep.hpp"
 #include "meep_internals.hpp"
 #include "config.h"
 
 #define RPR realnum *restrict
 
-/* These macros get into the guts of the PLOOP_OVER_VOL loops to
-   efficiently construct the index k into a PML sigma array.
-   Basically, k needs to increment by 2 for each increment of one of
-   LOOP's for-loops, starting at the appropriate corner of the grid_volume,
-   and these macros define the relevant strides etc. for each loop.
-   KSTRIDE_DEF defines the relevant strides etc. and goes outside the
-   LOOP, wheras KDEF defines the k index and goes inside the LOOP. */
-#define KSTRIDE_DEF(dsig, k, is, gv)                                                               \
-  const int k##0 = is.in_direction(dsig) - gv.little_corner().in_direction(dsig);              \
-  const int s##k##1 = gv.yucky_direction(0) == dsig ? 2 : 0;                                       \
-  const int s##k##2 = gv.yucky_direction(1) == dsig ? 2 : 0;                                       \
-  const int s##k##3 = gv.yucky_direction(2) == dsig ? 2 : 0
-#define KDEF(k, dsig)                                                                              \
-  const int k = ((k##0 + s##k##1 * loop_i1) + s##k##2 * loop_i2) + s##k##3 * loop_i3
-#define DEF_k KDEF(k, dsig)
-#define DEF_ku KDEF(ku, dsigu)
-#define DEF_kw KDEF(kw, dsigw)
-
 using namespace std;
 
 namespace meep {
@@ -63,13 +62,13 @@ namespace meep {
        df/dt = dfu/dt - sigma_u * f
    and fu replaces f in the equations above (fu += dt curl g etcetera).
 */
-void step_curl(RPR f, component c, const RPR g1, const RPR g2,
-               ptrdiff_t s1, ptrdiff_t s2, // strides for g1/g2 shift
-               const grid_volume &gv, const ivec is, const ivec ie,
-               realnum dtdx, direction dsig, const RPR sig, const RPR kap,
-               const RPR siginv, RPR fu, direction dsigu, const RPR sigu, const RPR kapu,
-               const RPR siginvu, realnum dt, const RPR cnd, const RPR cndinv, RPR fcnd) {
-  (void) c; // currently unused
+void step_curl(RPR f, component c, const RPR g1, const RPR g2, ptrdiff_t s1,
+               ptrdiff_t s2, // strides for g1/g2 shift
+               const grid_volume &gv, const ivec is, const ivec ie, realnum dtdx, direction dsig,
+               const RPR sig, const RPR kap, const RPR siginv, RPR fu, direction dsigu,
+               const RPR sigu, const RPR kapu, const RPR siginvu, realnum dt, const RPR cnd,
+               const RPR cndinv, RPR fcnd) {
+  (void)c;   // currently unused
   if (!g1) { // swap g1 and g2
     SWAP(const RPR, g1, g2);
     SWAP(ptrdiff_t, s1, s2);
@@ -253,10 +252,10 @@ void step_curl(RPR f, component c, const RPR g1, const RPR g2,
    and/or PML).  This is used in 2d calculations to add an exp(i beta z)
    time dependence, which gives an additional i \beta \hat{z} \times
    cross-product in the curl equations. */
-void step_beta(RPR f, component c, const RPR g, const grid_volume &gv, const ivec is, const ivec ie, realnum betadt,
-               direction dsig, const RPR siginv, RPR fu, direction dsigu, const RPR siginvu,
-               const RPR cndinv, RPR fcnd) {
-  (void) c; // currently unused
+void step_beta(RPR f, component c, const RPR g, const grid_volume &gv, const ivec is, const ivec ie,
+               realnum betadt, direction dsig, const RPR siginv, RPR fu, direction dsigu,
+               const RPR siginvu, const RPR cndinv, RPR fcnd) {
+  (void)c; // currently unused
   if (!g) return;
   if (dsig != NO_DIRECTION) { // PML in f update
     KSTRIDE_DEF(dsig, k, is, gv);
@@ -332,6 +331,206 @@ void step_beta(RPR f, component c, const RPR g, const grid_volume &gv, const ive
     }
   }
 }
+// allows fixed angle broadband simulations
+void step_bfast(RPR f, component c, const RPR g1, const RPR g2, ptrdiff_t s1,
+                ptrdiff_t s2, // strides for g1/g2 shift
+                const grid_volume &gv, const ivec is, const ivec ie, realnum dtdx, direction dsig,
+                const RPR sig, const RPR kap, const RPR siginv, RPR fu, direction dsigu,
+                const RPR sigu, const RPR kapu, const RPR siginvu, realnum dt, const RPR cnd,
+                const RPR cndinv, RPR fcnd, RPR F, realnum k1, realnum k2) {
+  (void)c;   // currently unused
+  if (!g1) { // swap g1 and g2
+    SWAP(const RPR, g1, g2);
+    SWAP(ptrdiff_t, s1, s2);
+    SWAP(realnum, k1, k2); // need to swap in cross product
+  }
+  if (dsig == NO_DIRECTION) {    // no PML in f update
+    if (dsigu == NO_DIRECTION) { // no fu update
+      if (cnd) {
+        if (g2) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            realnum F_prev = F[i];
+            F[i] = (k1 * (g1[i + s1] + g1[i]) - k2 * (g2[i + s2] + g2[i])) - F[i];
+            f[i] += (F[i] - F_prev) * cndinv[i];
+          }
+        }
+        else {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            realnum F_prev = F[i];
+            F[i] = k1 * (g1[i + s1] + g1[i]) - F[i];
+            f[i] += (F[i] - F_prev) * cndinv[i];
+          }
+        }
+      }
+      else { // no conductivity
+        if (g2) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            realnum F_prev = F[i];
+            F[i] = (k1 * (g1[i + s1] + g1[i]) - k2 * (g2[i + s2] + g2[i])) - F[i];
+            f[i] += (F[i] - F_prev);
+          }
+        }
+        else {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            realnum F_prev = F[i];
+            F[i] = k1 * (g1[i + s1] + g1[i]);
+            f[i] += (F[i] - F_prev);
+          }
+        }
+      }
+    }
+    else { // fu update, no PML in f update
+      KSTRIDE_DEF(dsigu, ku, is, gv);
+      if (cnd) {
+        if (g2) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            DEF_ku;
+            realnum df;
+            realnum F_prev = F[i];
+            F[i] = (k1 * (g1[i + s1] + g1[i]) - k2 * (g2[i + s2] + g2[i])) - F[i];
+            fu[i] += (df = (F[i] - F_prev) * cndinv[i]);
+            f[i] += siginvu[ku] * df;
+          }
+        }
+        else {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            DEF_ku;
+            realnum df;
+            realnum F_prev = F[i];
+            F[i] = k1 * (g1[i + s1] + g1[i]) - F[i];
+            fu[i] += (df = (F[i] - F_prev) * cndinv[i]);
+            f[i] += siginvu[ku] * df;
+          }
+        }
+      }
+      else { // no conductivity
+        if (g2) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            DEF_ku;
+            realnum df;
+            realnum F_prev = F[i];
+            F[i] = (k1 * (g1[i + s1] + g1[i]) - k2 * (g2[i + s2] + g2[i])) - F[i];
+            fu[i] += (df = (F[i] - F_prev));
+            f[i] += siginvu[ku] * df;
+          }
+        }
+        else {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            DEF_ku;
+            realnum df;
+            realnum F_prev = F[i];
+            F[i] = k1 * (g1[i + s1] + g1[i]) - F[i];
+            fu[i] += (df = (F[i] - F_prev));
+            f[i] += siginvu[ku] * df;
+          }
+        }
+      }
+    }
+  }
+  else { // PML in f update
+    KSTRIDE_DEF(dsig, k, is, gv);
+    if (dsigu == NO_DIRECTION) { // no fu update
+      if (cnd) {
+        realnum dt2 = dt * 0.5;
+        if (g2) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            DEF_k;
+            realnum F_prev = F[i];
+            F[i] = (k1 * (g1[i + s1] + g1[i]) - k2 * (g2[i + s2] + g2[i])) - F[i];
+            realnum dfcnd = (F[i] - F_prev) * cndinv[i];
+            fcnd[i] += dfcnd;
+            f[i] += dfcnd * siginv[k];
+          }
+        }
+        else {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            DEF_k;
+            realnum F_prev = F[i];
+            F[i] = k1 * (g1[i + s1] + g1[i]) - F[i];
+            realnum dfcnd = (F[i] - F_prev) * cndinv[i];
+            fcnd[i] += dfcnd;
+            f[i] += dfcnd * siginv[k];
+          }
+        }
+      }
+      else { // no conductivity (other than PML conductivity)
+        if (g2) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            DEF_k;
+            realnum F_prev = F[i];
+            F[i] = (k1 * (g1[i + s1] + g1[i]) - k2 * (g2[i + s2] + g2[i])) - F[i];
+            f[i] += (F[i] - F_prev) * siginv[k];
+          }
+        }
+        else {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            DEF_k;
+            realnum F_prev = F[i];
+            F[i] = k1 * (g1[i + s1] + g1[i]) - F[i];
+            f[i] += (F[i] - F_prev) * siginv[k];
+          }
+        }
+      }
+    }
+    else { // fu update + PML in f update
+      KSTRIDE_DEF(dsigu, ku, is, gv);
+      if (cnd) {
+        if (g2) {
+          //////////////////// MOST GENERAL CASE //////////////////////
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            DEF_k;
+            DEF_ku;
+            realnum df;
+            realnum F_prev = F[i];
+            F[i] = (k1 * (g1[i + s1] + g1[i]) - k2 * (g2[i + s2] + g2[i])) - F[i];
+            realnum dfcnd = (F[i] - F_prev) * cndinv[i];
+            fcnd[i] += dfcnd;
+            fu[i] += (df = dfcnd * siginv[k]);
+            f[i] += siginvu[ku] * df;
+          }
+          /////////////////////////////////////////////////////////////
+        }
+        else {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            DEF_k;
+            DEF_ku;
+            realnum df;
+            realnum F_prev = F[i];
+            F[i] = k1 * (g1[i + s1] + g1[i]) - F[i];
+            realnum dfcnd = (F[i] - F_prev) * cndinv[i];
+            fcnd[i] += dfcnd;
+            fu[i] += (df = dfcnd * siginv[k]);
+            f[i] += siginvu[ku] * df;
+          }
+        }
+      }
+      else { // no conductivity (other than PML conductivity)
+        if (g2) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            DEF_k;
+            DEF_ku;
+            realnum df;
+            realnum F_prev = F[i];
+            F[i] = (k1 * (g1[i + s1] + g1[i]) - k2 * (g2[i + s2] + g2[i])) - F[i];
+            fu[i] += (df = (F[i] - F_prev) * siginv[k]);
+            f[i] += siginvu[ku] * df;
+          }
+        }
+        else {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            DEF_k;
+            DEF_ku;
+            realnum df;
+            realnum F_prev = F[i];
+            F[i] = k1 * (g1[i + s1] + g1[i]) - F[i];
+            fu[i] += (df = (F[i] - F_prev) * siginv[k]);
+            f[i] += siginvu[ku] * df;
+          }
+        }
+      }
+    }
+  }
+}
 
 /* Given Dsqr = |D|^2 and Di = component of D, compute the factor f so
    that Ei = chi1inv * f * Di.   In principle, this would involve solving
@@ -364,11 +563,11 @@ inline realnum calc_nonlinear_u(const realnum Dsqr, const realnum Di, const real
 
 */
 
-void step_update_EDHB(RPR f, component fc, const grid_volume &gv, const ivec is, const ivec ie, const RPR g, const RPR g1,
-                      const RPR g2, const RPR u, const RPR u1, const RPR u2, ptrdiff_t s,
-                      ptrdiff_t s1, ptrdiff_t s2, const RPR chi2, const RPR chi3, RPR fw,
-                      direction dsigw, const RPR sigw, const RPR kapw) {
-  (void) fc; // currently unused
+void step_update_EDHB(RPR f, component fc, const grid_volume &gv, const ivec is, const ivec ie,
+                      const RPR g, const RPR g1, const RPR g2, const RPR u, const RPR u1,
+                      const RPR u2, ptrdiff_t s, ptrdiff_t s1, ptrdiff_t s2, const RPR chi2,
+                      const RPR chi3, RPR fw, direction dsigw, const RPR sigw, const RPR kapw) {
+  (void)fc; // currently unused
   if (!f) return;
 
   if ((!g1 && g2) || (g1 && g2 && !u1 && u2)) { /* swap g1 and g2 */
@@ -468,9 +667,7 @@ void step_update_EDHB(RPR f, component fc, const grid_volume &gv, const ivec is,
             f[i] += (kapwkw + sigwkw) * fw[i] - (kapwkw - sigwkw) * fwprev;
           }
         }
-        else if (g2) {
-          meep::abort("bug - didn't swap off-diagonal terms!?");
-        }
+        else if (g2) { meep::abort("bug - didn't swap off-diagonal terms!?"); }
         else {
           PLOOP_OVER_IVECS(gv, is, ie, i) {
             realnum gs = g[i];
@@ -565,9 +762,7 @@ void step_update_EDHB(RPR f, component fc, const grid_volume &gv, const ivec is,
                    calc_nonlinear_u(gs * gs + 0.0625 * (g1s * g1s), gs, us, chi2[i], chi3[i]);
           }
         }
-        else if (g2) {
-          meep::abort("bug - didn't swap off-diagonal terms!?");
-        }
+        else if (g2) { meep::abort("bug - didn't swap off-diagonal terms!?"); }
         else {
           PLOOP_OVER_IVECS(gv, is, ie, i) {
             realnum gs = g[i];
