@@ -1,4 +1,4 @@
-/* Copyright (C) 2005-2021 Massachusetts Institute of Technology.
+/* Copyright (C) 2005-2024 Massachusetts Institute of Technology
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -45,9 +45,12 @@ void harmonics(double freq, double chi2, double chi3, double J, double &A2, doub
   f.add_point_source(Ex, src, vec(-0.5 * sz + dpml), J);
 
   vec fpt(0.5 * sz - dpml - 0.5);
-  dft_flux d1 = f.add_dft_flux(Z, volume(fpt), freq, freq, 1);
-  dft_flux d2 = f.add_dft_flux(Z, volume(fpt), 2 * freq, 2 * freq, 1);
-  dft_flux d3 = f.add_dft_flux(Z, volume(fpt), 3 * freq, 3 * freq, 1);
+  dft_flux d1 =
+      f.add_dft_flux(Z, volume(fpt), freq, freq, 1, true, true, 1 /* decimation_factor */);
+  dft_flux d2 =
+      f.add_dft_flux(Z, volume(fpt), 2 * freq, 2 * freq, 1, true, true, 1 /* decimation_factor */);
+  dft_flux d3 =
+      f.add_dft_flux(Z, volume(fpt), 3 * freq, 3 * freq, 1, true, true, 1 /* decimation_factor */);
 
   double emax = 0;
 
@@ -98,10 +101,15 @@ int main(int argc, char **argv) {
 
   double a2, a3, a2_2, a3_2;
 
-  double thresh = sizeof(realnum) == sizeof(float) ? 1e-4 : 1e-5;
+  double thresh = sizeof(realnum) == sizeof(float) ? 1e-3 : 1e-5;
   harmonics(freq, 0.27e-4, 1e-4, 1.0, a2, a3);
   if (different(a2, 9.80330e-07, thresh, "2nd harmonic mismatches known val")) return 1;
-  if (different(a3, 9.97747e-07, thresh, "3rd harmonic mismatches known val")) return 1;
+  if (sizeof(realnum) == sizeof(float)) {
+    if (different(a3, 9.99349e-07, thresh, "3rd harmonic mismatches known val")) return 1;
+  }
+  else {
+    if (different(a3, 9.97747e-07, thresh, "3rd harmonic mismatches known val")) return 1;
+  }
 
   harmonics(freq, 0.54e-4, 2e-4, 1.0, a2_2, a3_2);
   master_printf("doubling chi2, chi3 = %g x 2nd harmonic, %g x 3rd\n", a2_2 / a2, a3_2 / a3);
@@ -121,7 +129,13 @@ int main(int argc, char **argv) {
   }
 
   harmonics(freq, 0.0, 1e-4, 1.0, a2_2, a3_2);
-  if (different(a3, a3_2, 1e-3, "chi2 has too big effect on 3rd harmonic")) return 1;
+  if (sizeof(realnum) == sizeof(float)) {
+    if (different(a3, a3_2, 0.0017, "chi2 has too big effect on 3rd harmonic")) return 1;
+  }
+  else {
+    if (different(a3, a3_2, 0.001, "chi2 has too big effect on 3rd harmonic")) return 1;
+  }
+
   if (a2_2 / a2 > 1e-5) {
     master_printf("error: too much 2nd harmonic without chi3\n");
     return 1;
